"""
NASA Docket No. GSC-19,469-1, and identified as "Astrophysics Cross-Observatory
Science Support (ACROSS) System

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software distributed
under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.“

The copyright notice to be included in the software is as follows:

Copyright © 2025 United States Government as represented by the Administrator
of the National Aeronautics and Space Administration and The Penn State
Research Foundation.  All rights reserved. This software is licensed under the
Apache 2.0 License.
"""

__path__ = __import__("pkgutil").extend_path(__path__, __name__)
